/*
 * Kosmos.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package hu.midori.kosmos.server.svn;

import hu.midori.kosmos.model.ScmRepositoryChange;
import hu.midori.kosmos.model.SvnRepository;
import hu.midori.kosmos.model.util.ModelUtils;
import hu.midori.kosmos.protocol.SvnService;
import hu.midori.kosmos.server.AbstractKosmosService;
import hu.midori.kosmos.server.util.ChartUtils;
import hu.midori.kosmos.server.util.UrlUtils;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jfree.data.time.TimeSeriesDataItem;
import org.jfree.data.time.Week;
import org.tmatesoft.svn.core.SVNDirEntry;
import org.tmatesoft.svn.core.SVNException;
import org.tmatesoft.svn.core.SVNLogEntry;
import org.tmatesoft.svn.core.SVNLogEntryPath;
import org.tmatesoft.svn.core.SVNNodeKind;
import org.tmatesoft.svn.core.SVNURL;
import org.tmatesoft.svn.core.internal.io.dav.DAVRepositoryFactory;
import org.tmatesoft.svn.core.internal.io.svn.SVNRepositoryFactoryImpl;
import org.tmatesoft.svn.core.io.SVNRepository;
import org.tmatesoft.svn.core.io.SVNRepositoryFactory;
import org.tmatesoft.svn.core.wc.SVNWCUtil;

/**
 * Implementation of the <i>Subversion</i> service.
 * <p>
 * Uses the <a href="http://tmate.org/svn/">JavaSVN library</a>
 * to access the repository.
 *
 * @author <a href="mailto:aron.gombas@midori.hu">Aron Gombas</a>
 * @version $Id$
 */
public class SvnServiceImpl extends AbstractKosmosService implements SvnService {
    private final static Log log = LogFactory.getLog(SvnServiceImpl.class);

    /** This "level 2 cache" stores the latest revision of repositories already processed. */
    private Map<String, SvnRepository> cacheLevel2 = new HashMap<String, SvnRepository>();

    @SuppressWarnings("unchecked")
    public List<SvnRepository> getRepositories(String url) {
        log.debug("Returning repositories...");

        List<SvnRepository> repositories = new ArrayList<SvnRepository>();
        updateTimestamp();

        SvnRepository cachedRepo = lookupLevel2Cache(url);
        if(cachedRepo != null) {
            log.info(String.format("\"%s\" revision didn't change, returning from L2 cache...", url));
            repositories.add(cachedRepo);
            return repositories;
        }

        beginStoreSession();

        try {
            // get repository info
            SVNRepository repo = connect(url);

            SvnRepositoryLogStats logStats = analyzeLog(repo);
            SvnRepositoryContentStats contentStats = analyzeContent(repo);
            generateCharts(repo, logStats, contentStats);

            disconnect(repo);

            // save
            SvnRepository repository = new SvnRepository(UrlUtils.removeUserInfoFromUrl(repo.getLocation().toString()), repo.getLatestRevision(),
                logStats.createdDate, trimList(logStats.commits),
                logStats.commitsTotal, logStats.commitsToday, logStats.commitsLast7Days, logStats.commitsLast31Days,
                logStats.committersTotal.size(), logStats.committersToday.size(), logStats.committersLast7Days.size(), logStats.committersLast31Days.size(),
                contentStats.filesPerFileTypeChartLocator, logStats.commitsPerAuthorChartLocator, logStats.commitsPerFileChartLocator, logStats.commitsPerWeekChartLocator, logStats.repoEntriesPerWeekChartLocator,
                contentStats.dirs, contentStats.files, contentStats.totalSize,
                trimList(contentStats.filesPerFileType), trimList(logStats.commitsPerAuthor), trimList(logStats.commitsPerFile));
            log.debug(String.format("Adding repository \"%s\"...", repository));
            repositories.add(repository);

            cacheLevel2.put(url, repository);
        } catch (Exception ex) {
            log.error("Unable to process the SVN repo", ex);
        }

        endStoreSession();

        return repositories;
    }

    /**
     * Connects to the given public or secure SVN URL.
     * @return the repository after a successful connection.
     */
    protected SVNRepository connect(String url) throws MalformedURLException, SVNException {
        // setup JavaSVN
        DAVRepositoryFactory.setup();
        SVNRepositoryFactoryImpl.setup();

        // connect to repository
        SVNURL svnUrl = SVNURL.parseURIEncoded(url);
        SVNRepository repo = SVNRepositoryFactory.create(svnUrl);

        // authenticate if repository is secured
        String userInfo[] = UrlUtils.parseUserInfoFromUrl(svnUrl.toString());
        if(userInfo != null) {
            String userName = userInfo[0];
            String password = userInfo[1];

            log.debug(String.format("Signing in as \"%s\" \"%s\"...", userName, password));
            repo.setAuthenticationManager(SVNWCUtil.createDefaultAuthenticationManager(userName, password));
        }

        return repo;
    }

    /** Disconnects from the given repository. */
    protected void disconnect(SVNRepository repo) {
        // do nothing
    }

    /**
     * Returns the repository from the "level 2 cache" if the latest revision
     * didn't change.
     */
    protected SvnRepository lookupLevel2Cache(String url) {
        try {
            SvnRepository oldRepo = cacheLevel2.get(url);
            if(oldRepo == null)
                return null;

            SVNRepository repo = connect(url);
            log.debug(String.format("Current revision: %s, level 2 cached revision: %s", repo.getLatestRevision(), oldRepo.getRevision()));
            boolean isUptodate = (repo.getLatestRevision() == oldRepo.getRevision());
            disconnect(repo);

            return isUptodate ? oldRepo : null;
        } catch(Exception ex) {
            log.error("Unable to get the latest revision", ex);
            return null;
        }
    }

    /** Analyzes the log of the repo and returns its stats. */
    @SuppressWarnings("unchecked")
    protected SvnRepositoryLogStats analyzeLog(SVNRepository repository) throws SVNException {
        SvnRepositoryLogStats stats = new SvnRepositoryLogStats();

        log.info(String.format("Analyzing log of \"%s\" repository...", repository.getLocation().toString()));
        Collection<SVNLogEntry> logEntries = (Collection<SVNLogEntry>)repository.log(new String[] {""}, null, 0, repository.getLatestRevision(), true, false);
        log.info(String.format("%d log-entries received.", logEntries.size()));

        // analyze
        long now = new Date().getTime();
        Map<String,Integer> commitsPerAuthorMap = new TreeMap<String,Integer>();
        Map<String,Integer> commitsPerFileMap = new TreeMap<String,Integer>();
        for(SVNLogEntry logEntry : logEntries) {
            // skip invalid entries
            if(logEntry.getAuthor() == null) {
                log.error("Skipping invalid entry (no author)");
                continue;
            }
            if(logEntry.getDate() == null) {
                log.error("Skipping invalid entry (no author)");
                continue;
            }

            // count commits-over-period and author-over-period
            long ageInMillis = now - logEntry.getDate().getTime();

            stats.commitsTotal++;
            stats.committersTotal.add(logEntry.getAuthor());
            if(ageInMillis < ModelUtils.MILLIS_PER_DAY) {
                stats.commitsToday++;
                stats.committersToday.add(logEntry.getAuthor());
            }
            if(ageInMillis < ModelUtils.MILLIS_PER_WEEK) {
                stats.commitsLast7Days++;
                stats.committersLast7Days.add(logEntry.getAuthor());
            }
            if(ageInMillis < ModelUtils.MILLIS_PER_MONTH) {
                stats.commitsLast31Days++;
                stats.committersLast31Days.add(logEntry.getAuthor());
            }

            // count commits-per-author
            ChartUtils.incrementCountMap(commitsPerAuthorMap, logEntry.getAuthor());

            // count commits-per-week
            Week commitWeek = new Week(logEntry.getDate());
            TimeSeriesDataItem commitsPerCurrentWeek = stats.commitsPerWeek.getDataItem(commitWeek);
            if(commitsPerCurrentWeek != null)
                stats.commitsPerWeek.addOrUpdate(commitWeek, commitsPerCurrentWeek.getValue().intValue() + 1);
            else
                stats.commitsPerWeek.add(commitWeek, 1);

            // count file-dependent stats
            Map changedPaths = logEntry.getChangedPaths();
            int addedFiles = 0;
            int deletedFiles = 0;
            int modifiedFiles = 0;
            int replacedFiles = 0;
            if(changedPaths.size() > 0) {
                for(Iterator it = changedPaths.keySet().iterator(); it.hasNext();) {
                    SVNLogEntryPath logEntryPath = (SVNLogEntryPath)changedPaths.get(it.next());
                    char type = logEntryPath.getType();

                    // update repo entry counters
                    int repoEntryAdder = 0;
                    switch(type) {
                        case 'A': addedFiles++; repoEntryAdder = 1; break; // 'add'
                        case 'D': deletedFiles++; repoEntryAdder = -1; break; // 'delete'
                        case 'M': modifiedFiles++; break; // 'modify'
                        case 'R': replacedFiles++; break; // 'replace'
                        default:  throw new IllegalArgumentException(String.format("Unknown commit type '%c'", type));
                    }

                    // count commits-per-file
                    ChartUtils.incrementCountMap(commitsPerFileMap, logEntryPath.getPath());

                    // count files-per-week
                    TimeSeriesDataItem repoEntriesPerCurrentWeek = stats.repoEntriesPerWeek.getDataItem(commitWeek);
                    if(repoEntriesPerCurrentWeek != null)
                        stats.repoEntriesPerWeek.addOrUpdate(commitWeek, repoEntriesPerCurrentWeek.getValue().intValue() + repoEntryAdder);
                    else
                        stats.repoEntriesPerWeek.add(commitWeek, repoEntryAdder);
                }
            }

            // save basic stats
            if (stats.createdDate == null)
                stats.createdDate = logEntry.getDate();

            stats.commits.add(new ScmRepositoryChange(logEntry.getRevision(), logEntry.getDate(), logEntry.getAuthor(), logEntry.getMessage(), addedFiles, deletedFiles, modifiedFiles, replacedFiles));
        }

        // validate
        stats.commitsPerAuthor = ChartUtils.countMapToSortedList(commitsPerAuthorMap);
        stats.commitsPerFile = ChartUtils.countMapToSortedList(commitsPerFileMap);
        stats.validate();

        return stats;
    }

    /** Analyzes the repository content and returns its stats. */
    @SuppressWarnings("unchecked")
    protected SvnRepositoryContentStats analyzeContent(SVNRepository repository) throws SVNException {
        SvnRepositoryContentStats stats = new SvnRepositoryContentStats();
        Map<String,Integer> filesPerFileTypeMap = new TreeMap<String,Integer>();

        log.info(String.format("Analyzing content of \"%s\" repository...", repository.getLocation().toString()));
        stats.dirs = 1; // starting root dir
        traverseRepository(repository, "", stats, filesPerFileTypeMap);
        log.info(String.format("%d repo-entries processed.", stats.dirs + stats.files));

        // validate
        stats.filesPerFileType = ChartUtils.countMapToSortedList(filesPerFileTypeMap);

        return stats;
    }

    /** Recursively traverses the given SVN repository tree and updates the stats passed. */
    @SuppressWarnings("unchecked")
    protected void traverseRepository(SVNRepository repository, String path, SvnRepositoryContentStats stats, Map<String,Integer> filesPerFileTypeMap) throws SVNException {
        Collection<SVNDirEntry> entries = (Collection<SVNDirEntry>)repository.getDir(path, -1, null, (Collection)null);

        for(SVNDirEntry entry : entries) {
            log.debug(String.format("Traversing \"/%s%s\"...", path.equals("") ? "" : path + "/", entry.getName()));

            if (entry.getKind() == SVNNodeKind.DIR) {
                // update stats
                stats.dirs++;

                // traverse recursively
                traverseRepository(repository, path.equals("") ? entry.getName() : path + "/" + entry.getName(), stats, filesPerFileTypeMap);
            } else if (entry.getKind() == SVNNodeKind.FILE){
                // update stats
                stats.files++;
                stats.totalSize += entry.getSize();

                // count file-types
                String fileType = StringUtils.substringAfterLast(entry.getName(), ".");
                if(!StringUtils.isEmpty(fileType))
                    ChartUtils.incrementCountMap(filesPerFileTypeMap, fileType.toLowerCase());
            }
        }
    }

    /** Generates and saves the chart images on the server, and sets the URLs to the stats. */
    protected void generateCharts(SVNRepository repository, SvnRepositoryLogStats logStats, SvnRepositoryContentStats contentStats) throws IOException {
        String prefix = UrlUtils.removeUserInfoFromUrl(repository.getLocation().toString());

        // generate charts
        contentStats.filesPerFileTypeChartLocator = storeChart(prefix + "_files_per_file_type", ChartUtils.generatePieChart(ChartUtils.collectionToPieDataset(contentStats.filesPerFileType, ChartUtils.MAX_ITEMS)));

        logStats.commitsPerAuthorChartLocator = storeChart(prefix + "_commits_per_author", ChartUtils.generatePieChart(ChartUtils.collectionToPieDataset(logStats.commitsPerAuthor, ChartUtils.MAX_ITEMS)));
        logStats.commitsPerFileChartLocator = storeChart(prefix + "_commits_per_file", ChartUtils.generatePieChart(ChartUtils.collectionToPieDataset(logStats.commitsPerFile, ChartUtils.MAX_ITEMS)));
        logStats.commitsPerWeekChartLocator = storeChart(prefix + "_commits_per_week", ChartUtils.generateTimeBarChart(logStats.commitsPerWeek));
        logStats.repoEntriesPerWeekChartLocator = storeChart(prefix + "_files_per_week", ChartUtils.generateTimeLineChart(logStats.repoEntriesPerWeek));
    }
}
